var INDEX_UUID=0, INDEX_TYPE=1, INDEX_NAME=2, INDEX_URL=3, INDEX_DATE=4, INDEX_FOLDER=5, INDEX_PASSWORD_STRENGTH=6, INDEX_TRASHED=7;
var TYPE_WEBFORMS='webforms.WebForm', TYPE_FOLDERS='system.folder.Regular', TYPE_NOTES='securenotes.SecureNote', TYPE_IDENTITIES='identities.Identity', TYPE_PASSWORDS='passwords.Password', TYPE_WALLET='wallet', TYPE_SOFTWARE_LICENSES='wallet.computer.License', TYPE_TRASHED='trashed';
var ERROR_BAD_DECRYPT = "Decryption failed", ERROR_INVALID_JSON = "Decryption passed but JSON was invalid", ERROR_OK = "OK";

var Keychain = Class.create({
	AUTOLOCK_LENGTH: 5 * 60 * 1000, autoLogoutTime: null,
	contents: {webforms:[], wallet:[], notes:[], identities:[], passwords:[], folders:[]},
	encryptionKeys: null, _all: null, masterPassword: null,
	encryptionKeysLoaded: false,
	
	initialize: function() {
	},
	
	setEncryptionKeys: function(keys) {
		this.encryptionKeys = keys;
		this.encryptionKeysLoaded = true;
	},
	
	verifyPassword: function(password) {
		GibberishAES.size(128);
		this.encryptionKeys.decryptedKeys = {};

		var key = this.decryptEncryptionKey("SL5", password);
		if (!key) return false;

		this.masterPassword = password;
		return true;
	},

	decryptEncryptionKey: function(sl, password) {
		
		for (var i=0; i < this.encryptionKeys["list"].length; i++) {
			var item = this.encryptionKeys["list"][i];

			if (item['identifier'] == this.encryptionKeys[sl]) {
				var decryptedKey = GibberishAES.decryptUsingPBKDF2(item["data"], password);
				if (!decryptedKey) return null;

				var verification = GibberishAES.decryptBase64UsingKey(item["validation"],GibberishAES.s2a(decryptedKey));
				if (verification != decryptedKey) return null;

				this.encryptionKeys.decryptedKeys[sl] = decryptedKey;
				return decryptedKey;
			}
		}

		return null;
	},

	keyForItem: function(item) {
		if (item.securityLevel == null) {
			return this.encryptionKeys.decryptedKeys["SL5"]
		}
		
		var key = this.encryptionKeys.decryptedKeys[item.securityLevel];
		if (!key) {
			key = this.decryptEncryptionKey(item.securityLevel, this.masterPassword);
		}
		return key;
	},

	rescheduleAutoLogout: function () {
		this.autoLogoutTime = new Date().getTime() + this.AUTOLOCK_LENGTH;
	},
	
	setContents: function(items) {
		_all = items;
		this.contents[TYPE_WEBFORMS] = this._keychainItemsOfType(TYPE_WEBFORMS, false);
		this.contents[TYPE_NOTES] = this._keychainItemsOfType(TYPE_NOTES, false);
		this.contents[TYPE_WALLET] = this._keychainItemsOfType(TYPE_WALLET, false);
		this.contents[TYPE_PASSWORDS] = this._keychainItemsOfType(TYPE_PASSWORDS, false);
		this.contents[TYPE_IDENTITIES] = this._keychainItemsOfType(TYPE_IDENTITIES, false);
		this.contents[TYPE_SOFTWARE_LICENSES] = this._keychainItemsOfType(TYPE_SOFTWARE_LICENSES, false);
		this.contents['folders'] = this._keychainItemsOfType(TYPE_FOLDERS, false);
		this.contents[TYPE_TRASHED] = this._keychainItemsOfType("", true);
		
		this.rescheduleAutoLogout();
// TODO: Timeout is not working properly....?
//		window.setTimeout(this._autoLogout, 750);
	},
	
	itemsOfType : function(name) {
		return this.contents[name];
	},
	
	itemWithUuid : function(uuid) {
		return _all.find(function(item){return item[INDEX_UUID] == this}, uuid);
	},
	
	_keychainItemsOfType : function(type, trashed) {

		// We're not compatible w/ 2.x. Warning will save me an hour next time this happens.
		if (_all[0][INDEX_TRASHED]==undefined) {
			alert("Your 1Password keychain was last used with version 2 and 1PasswordAnywhere requires version 3. Launch 1Password 3 and change something to force your keychain to be updated and then try again.");
			throw "Version 3 keychain required.";
		}
		
		var result = [];
		var rawList = _all.findAll(function(item) {return item[INDEX_TYPE].startsWith(this)}, type);
		var trash_state = trashed ? "Y" : "N";
		for (var i=0; i < rawList.length; i++) {
			
			var item = rawList[i];
			if (type == TYPE_WALLET && item[INDEX_TYPE] == TYPE_SOFTWARE_LICENSES) continue;
			
			if (item[INDEX_TRASHED] == trash_state) {
				result.push(new KeychainItemOverview(rawList[i]));
			}
		}
		return result;
	},
	
	_autoLogout: function() {
		var now = new Date();

		if (now.getTime() < this.autoLogoutTime) {
			window.setTimeout(this._autoLogout, 750);
			return;
		}
		logout(true);
	},
	
	lock: function() {
		if (this.encryptionKeys) {
			this.encryptionKeys.decryptedKeys = null;
		}
		this.masterPassword = null;
		this.encryptionKeysLoaded = false;
	}
});

var KeychainItemOverview = Class.create({
	uuid: null, type: null, title: null, domain: null, updatedAt: null, updatedAtMs: null, trashed: null,

	initialize: function(data) {
		this.uuid = data[INDEX_UUID];
		this.type = data[INDEX_TYPE];
		this.title = data[INDEX_NAME];
		this.domain = data[INDEX_URL];
		this.updatedAtMs = data[INDEX_DATE] * 1000;
		this.trashed = data[INDEX_TRASHED];
		
		var date = new Date();
		date.setTime(this.updatedAtMs);
		this.updatedAt = date.format("mmm d, yyyy, h:MM:ss TT");
	},
	
});

var KeychainItem = Class.create({
	uuid: null, type: null, title: null, domain: null, updatedAt: null, updatedAtMs: null,
	folderUuid: null, folderName: null, data: null, 
	decrypted: false, encrypted_contents: null, decrypted_secure_contents: null,
	updatedAtMs: null,

	initialize: function(data) {
		this.data = data;
		this.folderUuid = data.folderUuid;
		this.encrypted_contents = data.encrypted;
		this.decrypted = false, this.decrypted_secure_contents = null;
		
		this.uuid = data.uuid;
		this.type = data.typeName;
		this.title = data.title;
		this.domain = data.location;
		this.securityLevel = data.openContents.securityLevel;
		this.updatedAtMs = data.updatedAt * 1000;
		
		folderName = keychain.itemWithUuid(this.folderUuid);
		
		if (!folderName) folderName = "None";
	},
	
	decrypt: function() {
		GibberishAES.size(128);
		var plainText = null;
		var key = keychain.keyForItem(this);
		plainText = GibberishAES.decryptBase64UsingKey(this.encrypted_contents, GibberishAES.s2a(key));

		if (!plainText) {
			return ERROR_BAD_DECRYPT;
		}

		var pt = '(' + plainText + ');';
		try { this.decrypted_secure_contents = eval(pt); } 
		catch (e) { return ERROR_INVALID_JSON; }
		
		return ERROR_OK;
	},
	
	isWebForm: function()  {
		return this.data.typeName.indexOf(TYPE_WEBFORMS) == 0;
	},
	
	asHTML: function() {
		switch(this.type) {
			case TYPE_WEBFORMS: return this.webformHTML(); break;
			case TYPE_NOTES: return this.secureNoteHTML(); break;
			case TYPE_SOFTWARE_LICENSES: return this.walletHTML(); break;
			case TYPE_IDENTITIES: return this.identityHTML(); break;
			case TYPE_PASSWORDS: return this.passwordHTML();
		}
		
		if (this.type.startsWith(TYPE_WALLET)) {
			return this.walletHTML();
		}
		
		return "No asHTML defined for " + this.type+ " uuid:" + this.uuid;
	},
	
	passwordHTML: function() {
		var r = "<table><tr><th></th>";
		r += "<th><h1>" + this.title + "</h1></th></tr>";
		r += this.spacerTR();
		r += this.notesTR();
		r += this.spacerTR();		
		r += "<tr><td class='label'>password</td><td>" + showAndHideConcealedField(this.uuid, this.decrypted_secure_contents.password, true) + "</td></tr></table>";
		return r;
	},

	identityHTML: function() {
		var r = "<table><tr><th></th>";
		r += "<th><h1>" + this.title + "</h1></th></tr>";
		r += this.spacerTR();
		r += this.notesTR();
		r += this.spacerTR();
		r += this.walletFields();
		r += "</table>";
		return r;
	},
	
	walletHTML: function() {
		var r = "<table><tr><th></th>";
		r += "<th><h1>" + this.title + "</h1></th></tr>";
		r += this.spacerTR();
		r += this.notesTR();
		r += this.spacerTR();
		r += this.walletFields();
		r += "</table>";
		return r;
	},
	
	walletFields: function() {
		var entryValues = "";
		for (var key in this.decrypted_secure_contents)
		{
			if (key == 'notesPlain') continue;
			var v = this.decrypted_secure_contents[key];
			entryValues += "<tr><td class='label'>" + key.escapeHTML() + "</td><td>";

			try {entryValues += v.escapeHTML();}
			catch(e) {entryValues += v;}
			entryValues += "</td></tr>";
		}
		return entryValues;
	},
	
	secureNoteHTML: function() {
		var r = "<table><tr colspan='2'>";
		r += "<th><h1>" + this.title + "</h1></th></tr>";

		var secure = this.decrypted_secure_contents;
		if (!secure.notesPlain || secure.notesPlain.length == 0) return "";
		
		return r + "<tr><td colspan='2'>" + convertNewLines(secure.notesPlain.escapeHTML()) + "</td></tr></table>";
	},
	
	webformHTML: function() {	
		var r = "<table><tr><th></th>";
		r += "<th><h1>" + this.title + "</h1>" + this.domain + "</th></tr>";

		r += "<tr><td class='label'>Username</td><td>" + this.loginUsername() + "</td></tr>";
	    r += "<tr><td class='label'>Password</td><td>" + showAndHideConcealedField(this.uuid+'-password', this.loginPassword(), true) + "</td></tr>";

		r += this.spacerTR();
		if (devmode) {
			r += "<tr><td class='label'>UUID</td><td>" + this.uuid + "</td></tr>";
		}
		r += this.notesTR();
		r += this.spacerTR();
		r += this.loginFields();
		r += "</table>";
		return r;
	},
	
	findFieldWithDesignation: function(designation) {
		var count = this.decrypted_secure_contents.fields.length;
		for (var i = 0; i < count; ++i)
		{
			v = this.decrypted_secure_contents.fields[i];
			if (v.designation == designation) {
				return v.value;
			}
		}
		return null;
	},
	
	loginUsername: function() {
		var r = this.findFieldWithDesignation('username');
		if (r) return r;
		return "no field was designated as username";
	},
	
	loginPassword: function() {
		var r = this.findFieldWithDesignation('password');
		if (r) return r;
		return "no field was designated as password";
	},
	
	loginFields: function() {
		var entryValues = "";
		var count = this.decrypted_secure_contents.fields.length;
		for (var i = 0; i < count; ++i)
		{
			v = this.decrypted_secure_contents.fields[i];
			if (v.name == undefined || v.value == undefined) continue;
			entryValues += "<tr><td class='label'>" + v.name.escapeHTML() + "</td><td>";

			if (v.type == 'P' || v.designation == 'password') {
				var id = "concealedField-" + this.uuid + "-" + i;
				entryValues += showAndHideConcealedField(id, v.value, true);
			}
			else {
				entryValues += v.value.escapeHTML();
			}
			entryValues += "</td></tr>";
		}
		return entryValues;
	},
	
	spacerTR:function() {
		return "<tr><td class='spacer'></td></tr>";
	},
	
	notesTR:function() {
		var secure = this.decrypted_secure_contents;
		if (!secure.notesPlain || secure.notesPlain.length == 0) return "";
		
		return "<tr><td class='label'>Notes</td><td>" + convertNewLines(secure.notesPlain.escapeHTML()) + "</td></tr>";
	},
});

function encodeToHex(str){
    var r="";
    var e=str.length;
    var c=0;
    var h;
    while(c<e){
        h=str.charCodeAt(c++).toString(16);
        while(h.length<2) h="0"+h;
        r+=h;
    }
    return r.toUpperCase();
}